#pragma once
#include <vector>
#include <string>
#include <memory>
#include <cstdint>
#include "ast/type_info.hpp"
#include "ast/context.hpp"
#include "symbol_table/symbol_table.hpp"
#include "token/token.hpp"


// LLVM Code Generator Classes
#include "llvm/IR/LLVMContext.h"
#include "llvm/IR/IRBuilder.h"
#include "llvm/IR/Module.h"
#include "llvm/IR/Verifier.h"
#include "llvm/IR/ValueSymbolTable.h"

// LLVM Values Classes
#include "llvm/IR/Constants.h"
#include "llvm/IR/BasicBlock.h"
#include "llvm/IR/Function.h"

// LLVM Type Classes
#include "llvm/IR/Type.h"

// LLVM Utiliies
#include "llvm/Support/Casting.h"


namespace quant_c_ast {
    /* Forward declaration of classes */
    /* Base class for all nodes */
    class ASTNode;
    class NullNode;

    /* Node for the entire program */
    class TranslationUnitNode;

    /* Nodes for different kinds of declarations */
    class DeclNode;
    class DeclarationNode;
    class InitDeclaratorNode;
    /*
    class VariableDeclNode;
    class VarDeclNode;
    class ConstDeclNode;
    class ArrayDeclNode;
    */

    class FuncDefNode;
    class DeclaratorNode;
    class DirectDeclaratorNode;
    class IdentifierDeclaratorNode;
    class ParensDeclaratorNode;
    class ArrayDeclaratorNode;
    class FunctionDeclaratorNode;
    class DestructureDeclaratorNode;

    class AbstractDeclaratorNode;
    class NullAbstractNode;
    class DirectAbstractDeclaratorNode;
    class NullDirectAbstractNode;

    /* Nodes for type values */
    class TypeNode;
    class NullTypeNode;

    class TypeSpecNode;
    class NullTypeSpecNode;
    class IntTypeNode;
    class BoolTypeNode;
    class CharTypeNode;
    class VoidTypeNode;
    class ShortTypeNode;
    class LongTypeNode;
    class FloatTypeNode;
    class DoubleTypeNode;
    class SignedTypeNode;
    class UnsignedTypeNode;
    class CustomTypeNode;

    class StorageClassSpecNode;
    class TypeDefNode;
    class ExternNode;
    class StaticNode;
    class AutoNode;
    class RegisterNode;
    class QuantumNode;

    class TypeQualifierNode;
    class ConstNode;
    class VolatileNode;

    class PointerNode;

    /* TODO: Define the following type nodes */
    /*
    class StructTypeNode;
    class UnionTypeNode;
    class TypeFieldNode;
    class EnumNode;
    class EnumFieldNode;
    */
    class TypeNameNode;


    /* Nodes for function arguments */
    class ArgDeclNode;
    class ArrayArgDeclNode;
    class ValArgDeclNode;
    class NamedArgDeclNode;
    class AbstractArgDeclNode;
    class TypeArgDeclNode;


    /* Nodes for statements */
    class StmtNode;
    class NullStmtNode;
    class IfThenNode;
    class ForNode;
    class BreakNode;
    class ContinueNode;
    class ReturnNode;
    class ExprStmtNode;
    class BlockNode;
    class NullBlockNode;
    class LabelStmtNode;
    class SwitchNode;
    class CaseStmtNode;
    class WhileNode;
    class GotoNode;

    /* Nodes for expressions */
    class ExprNode;
    class NullExprNode;
    class BinaryOpNode;   //TODO: extend with other binary operations
    class UnaryOpNode;
    class ConditionalExprNode;
    class CastNode;
    class CallExprNode;
    class IdentNode;
    class NullIdentNode;
    class IndexExprNode;
    class StrLitNode;
    class IntLitNode;
    class FloatLitNode;
    class CharLitNode;
    class TrueNode;
    class FalseNode;
    /* TODO: Define the following expression nodes */
    class InitializerListNode;
    class PostfixExprNode;
    class AsgExprNode;
    class ConditionalExprNode;

    /* Class for operators nodes */
    class OperatorNode;
    class UnaryOperatorNode;
    class QuantumUnaryOperatorNode;
    class QuantumUnaryPostfixNode;
    class AssignmentOperatorNode;


    /* Class definitions */

    /* Generic Nodes */
    class ASTNode {
    // abstract superclass; only subclasses are actually created
        public:
            static void genIndent(int indent);

            static void mustBe(bool assertion, std::string errorMsg);
            static void mustNotBe(bool assertion, std::string errorMsg);

            static void storageMustBe(type_info::Type testType, type_info::StorageSpecVal requiredType, std::string errorMsg);
            static void storageMustBe(type_info::Type testType, std::vector<type_info::StorageSpecVal> requiredStorage, std::string errorMsg);

            static void qualityMustBe(type_info::Type testType, type_info::TypeQualityVal requiredQuality, std::string errorMsg);
            static void qualityMustBe(type_info::Type testType, std::vector<type_info::TypeQualityVal> requiredQuality, std::string errorMsg);

            static void typeMustBe(type_info::Type testType, type_info::TypeValue requiredType, std::string errorMsg);
            static void typeMustBe(type_info::Type testType, std::vector<type_info::TypeValue> requiredType, std::string errorMsg);

            static void storageMustNotBe(type_info::Type testType, type_info::Type requiredType, std::string errorMsg);
            static void storageMustNotBe(type_info::Type testType, std::vector<type_info::Type> requiredType, std::string errorMsg);

            static void qualityMustNotBe(type_info::Type testType, type_info::Type requiredType, std::string errorMsg);
            static void qualityMustNotBe(type_info::Type testType, std::vector<type_info::Type> requiredType, std::string errorMsg);

            static void typeMustNotBe(type_info::Type testType, type_info::TypeValue invalidType, std::string errorMsg);
            static void typeMustNotBe(type_info::Type testType, std::vector<type_info::TypeValue> invalidType, std::string errorMsg);

            static void typesMustBeEqual(type_info::Type type1, type_info::Type type2, std::string errorMsg);

            static void typesMustBeSimilar(type_info::Type type1, type_info::Type type2, std::vector<type_info::TypeValue> validType, std::string errorMsg);

            std::string error();

            enum ASTNodeKind {
                AST_NullNode,
                AST_TranslationUnitNode,
                AST_DeclNode,
                    AST_DeclarationNode,
                    AST_InitDeclaratorNode,
                    AST_FuncDefNode,
                    AST_DeclaratorNode,
                    AST_DirectDeclaratorNode,
                        AST_IdentifierDeclaratorNode,
                        AST_ParensDeclaratorNode,
                        AST_ArrayDeclaratorNode,
                        AST_FunctionDeclaratorNode,
                        AST_DestructureDeclaratorNode,
                    AST_AbstractDeclaratorNode,
                        AST_NullAbstractNode,
                    AST_DirectAbstractDeclaratorNode,
                        AST_NullDirectAbstractNode,
                AST_TypeNode,
                    AST_NullTypeNode,
                    AST_TypeSpecNode,
                        AST_NullTypeSpecNode,
                        AST_IntTypeNode,
                        AST_BoolTypeNode,
                        AST_CharTypeNode,
                        AST_VoidTypeNode,
                        AST_ShortTypeNode,
                        AST_LongTypeNode,
                        AST_FloatTypeNode,
                        AST_DoubleTypeNode,
                        AST_SignedTypeNode,
                        AST_UnsignedTypeNode,
                        AST_CustomTypeNode,
                    AST_StorageClassSpecNode,
                        AST_TypeDefNode,
                        AST_ExternNode,
                        AST_StaticNode,
                        AST_AutoNode,
                        AST_RegisterNode,
                        AST_QuantumNode,
                    AST_TypeQualifierNode,
                        AST_ConstNode,
                        AST_VolatileNode,
                    AST_PointerNode,
                    AST_TypeNameNode,
                AST_ArgDeclNode,
                    AST_ArrayArgDeclNode,
                    AST_ValArgDeclNode,
                    AST_NamedArgDeclNode,
                    AST_AbstractArgDeclNode,
                    AST_TypeArgDeclNode,
                AST_StmtNode,
                    AST_NullStmtNode,
                    AST_IfThenNode,
                    AST_ForNode,
                    AST_BreakNode,
                    AST_ContinueNode,
                    AST_ReturnNode,
                    AST_ExprStmtNode,
                    AST_BlockNode,
                    AST_NullBlockNode,
                    AST_LabelStmtNode,
                    AST_SwitchNode,
                    AST_CaseStmtNode,
                    AST_WhileNode,
                    AST_GotoNode,
                AST_ExprNode,
                    AST_NullExprNode,
                    AST_BinaryOpNode,
                    AST_UnaryOpNode,
                    AST_ConditionalExprNode,
                    AST_CastNode,
                    AST_CallExprNode,
                    AST_InitializerListNode,
                    AST_IdentNode,
                    AST_NullIdentNode,
                    AST_IndexExprNode,
                    AST_StrLitNode,
                    AST_IntLitNode,
                    AST_FloatLitNode,
                    AST_CharLitNode,
                    AST_TrueNode,
                    AST_FalseNode,
                    AST_PostfixExprNode,
                    AST_AsgExprNode,
                    AST_OperatorNode,
                        AST_UnaryOperatorNode,
                        AST_QuantumUnaryOperatorNode,
                        AST_QuantumUnaryPostfixNode,
                        AST_AssignmentOperatorNode
            };

            ASTNode(ASTNodeKind nodeKind);
            ASTNode(ASTNodeKind nodeKind, int line, int col);
            virtual bool isNull();
            virtual void unparse(int indent = 0);
            virtual void checkSemantics();
            virtual llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);
            virtual quant_c_ast::QuantumValue quantum_codegen(QuantumContext& context);
            virtual ASTNodeKind getNodeKind() const { return this->nodeKind; };
            // virtual llvm::Type* get_type();
            int get_line() { return this->linenum; };
            int get_col()  { return this->colnum;  };



            // Members
            static symbol_table::SymbolTable<symbol_table::SymbolInfo> semanticTable;
            static symbol_table::SymbolTable<symbol_table::SymbolValue> codegenTable;
            static int semanticsErrors; // Total semantics errors found
            static int qubitOffset;
            static int numQuantumFuncs;
            static bool firstQuantFunc;
            static std::string quilCode;
            static llvm::LLVMContext LLVMContext;
            static llvm::IRBuilder<> Builder;
            static std::unique_ptr<llvm::Module> LLVMModule;
            // static llvm::ValueSymbolTable NamedValues;
            static bool classof(const ASTNode * node);
            int linenum;
            int colnum;

        private:
            ASTNodeKind nodeKind;

    }; // class ASTNode

    class NullNode : public ASTNode {
        public:
            NullNode();
            bool isNull();
            void unparse(int indent = 0);
            static bool classof(const ASTNode * node);
    }; // class NullNode



    /* TypeNode & subclasses */
    class TypeNode : public ASTNode {
        public:
            TypeNode(ASTNodeKind nodeKind);
            TypeNode(ASTNodeKind nodeKind, int line, int col, type_info::Type type);

            static bool classof(const ASTNode * node);
            static std::shared_ptr<NullTypeNode> NULL_VAL;
        // Members
            type_info::Type type; // Used for type-checking -- the type of this typeNode
    }; // class TypeNode

    class NullTypeNode : public TypeNode {
        public:
            NullTypeNode();
            bool isNull();
            void unparse(int indent = 0);

            static bool classof(const ASTNode * node);
    }; // class NullTypeNode

    class TypeSpecNode : public TypeNode {
        public:
            TypeSpecNode(ASTNodeKind nodeKind);
            TypeSpecNode(ASTNodeKind nodeKind, int line, int col, type_info::Type type);

            static bool classof(const ASTNode * node);
            static std::shared_ptr<NullTypeSpecNode> NULL_VAL;
    }; // class TypeSpecNode

    class NullTypeSpecNode : public TypeSpecNode {
        public:
            NullTypeSpecNode();
            bool isNull();
            void unparse(int indent = 0);
            static bool classof(const ASTNode * node);
    };

    class IntTypeNode : public TypeSpecNode {
        public:
            IntTypeNode(int line, int col);
            void unparse(int indent = 0);
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
            llvm::Type * codegen_type();
    }; // class IntTypeNode

    class BoolTypeNode : public TypeSpecNode {
        public:
            BoolTypeNode(int line, int col);
            void unparse(int indent = 0);
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
            // llvm::Type* get_type();
    }; // class BoolTypeNode

    class CharTypeNode : public TypeSpecNode {
        public:
            CharTypeNode(int line, int col);
            void unparse(int indent = 0);
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
            // llvm::Type* get_type();
    }; // class CharTypeNode

    class VoidTypeNode : public TypeSpecNode {
        public:
            VoidTypeNode(int line, int col);
            void unparse(int indent = 0);
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
            // llvm::Type* get_type();
    }; // class VoidTypeNode

    class ShortTypeNode : public TypeSpecNode {
        public:
            ShortTypeNode(int line, int col);
            void unparse(int indent = 0);
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
            // llvm::Type* get_type();
    }; // class ShortTypeNode

    class LongTypeNode : public TypeSpecNode {
        public:
            LongTypeNode(int line, int col);
            void unparse(int indent = 0);
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
            // llvm::Type* get_type();
    }; // class LongTypeNode

    class FloatTypeNode : public TypeSpecNode {
        public:
            FloatTypeNode(int line, int col);
            void unparse(int indent = 0);
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
            // llvm::Type* get_type();
    }; // class FloatTypeNode

    class DoubleTypeNode : public TypeSpecNode {
        public:
            DoubleTypeNode(int line, int col);
            void unparse(int indent = 0);
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
            // llvm::Type* get_type();
    }; // class DoubleTypeNode

    class SignedTypeNode : public TypeSpecNode {
        public:
            SignedTypeNode(int line, int col);
            void unparse(int indent = 0);
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
            // llvm::Type* get_type();
    }; // class SignedTypeNode

    class UnsignedTypeNode : public TypeSpecNode {
        public:
            UnsignedTypeNode(int line, int col);
            void unparse(int indent = 0);
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
            // llvm::Type* get_type();
    }; // class UnsignedTypeNode

    class CustomTypeNode : public TypeSpecNode {
        public:
            CustomTypeNode(std::string type_name, int line, int col);
            void unparse(int indent = 0);
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
            // Members
            std::string type_name;
    }; // class CustomTypeNode


    class StorageClassSpecNode : public TypeNode {
        public:
            StorageClassSpecNode(ASTNodeKind nodeKind, int line, int col);

            static bool classof(const ASTNode * node);
    }; // class StorageClassSpecNode

    class TypeDefNode : public StorageClassSpecNode {
        public:
            TypeDefNode(int line, int col);
            void unparse(int indent = 0);
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
    }; // class TypeDefNode

    class ExternNode : public StorageClassSpecNode {
        public:
            ExternNode(int line, int col);
            void unparse(int indent = 0);
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
            // llvm::Type* get_type();
    }; // class ExternNode

    class StaticNode : public StorageClassSpecNode {
        public:
            StaticNode(int line, int col);
            void unparse(int indent = 0);
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
            // llvm::Type* get_type();
    }; // class StaticNode

    class AutoNode : public StorageClassSpecNode {
        public:
            AutoNode(int line, int col);
            void unparse(int indent = 0);
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
            // llvm::Type* get_type();
    }; // class AutoNode

    class RegisterNode : public StorageClassSpecNode {
        public:
            RegisterNode(int line, int col);
            void unparse(int indent = 0);
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
            // llvm::Type* get_type();
    }; // class RegisterNode

    class QuantumNode : public StorageClassSpecNode {
        public:
            QuantumNode(int line, int col);
            void unparse(int indent = 0);
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
    }; // class QuantumNode

    class TypeQualifierNode : public TypeNode {
        public:
            TypeQualifierNode(ASTNodeKind nodeKind, int line, int col);

            static bool classof(const ASTNode * node);
    }; // class TypeQualifierNode

    class ConstNode : public TypeQualifierNode {
        public:
            ConstNode(int line, int col);
            void unparse(int indent = 0);
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
            // llvm::Type* get_type();
    }; // class ConstNode

    class VolatileNode : public TypeQualifierNode {
        public:
            VolatileNode(int line, int col);
            void unparse(int indent = 0);
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
            // llvm::Type* get_type();
    }; // class VolatileNode

    class PointerNode : public TypeNode {
        public:
            typedef std::vector<std::shared_ptr<TypeQualifierNode>>
                    TypeQualifierList;
            PointerNode(TypeQualifierList qualifiers,
                        int line, int col);
            void unparse(int indent = 0);
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
        // Members
        TypeQualifierList qualifiers;
    }; // class PointerNode

    class TypeNameNode : public TypeNode {
        public:
            typedef std::vector<std::shared_ptr<TypeNode>> SpecQualList;
            TypeNameNode(SpecQualList spec_qual_list,
                    std::shared_ptr<AbstractDeclaratorNode> abstract_declarator,
                    int line, int col);
            void unparse(int indent = 0);
            type_info::Type generateType();
            void checkSemantics();

            static bool classof(const ASTNode * node);
        // Members
            SpecQualList spec_qual_list;
            std::shared_ptr<AbstractDeclaratorNode> abstract_declarator;
    };

    // end TypeNode subclasses



    /* ExprNode & subclasses */
    class ExprNode : public ASTNode {
        public:
            ExprNode(ASTNodeKind nodeKind);
            ExprNode(ASTNodeKind nodeKind, int line, int col);
            ExprNode(ASTNodeKind nodeKind, int line, int col, type_info::Type type, type_info::Kind kind);

            static bool classof(const ASTNode * node);
            static std::shared_ptr<NullExprNode> NULL_VAL;

            // Used for type-checking: the type of this node
            type_info::Type type;
            // Used for type-checking: the kind of this node
            type_info::Kind kind;
            bool isLvalue = false;
    }; // abstract class ExprNode

    class NullExprNode : public ExprNode {
        public:
            NullExprNode();
            bool isNull();
            void unparse(int indent = 0);

            static bool classof(const ASTNode * node);
    }; // class NullExprNode

    class IdentNode : public ExprNode {
        public:
            IdentNode();
            IdentNode(std::string idname, int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);
            quant_c_ast::QuantumValue quantum_codegen(QuantumContext& context);

            static bool classof(const ASTNode * node);
            static std::shared_ptr<NullIdentNode> NULL_VAL;

        // Members
            std::string idname;
            symbol_table::SymbolInfo id_info; // symbol table entry for this ident
            bool genAddr = false; // Used during codegen to change whether the expression is the variable's value or address
    }; // class IdentNode

    class BinaryOpNode : public ExprNode {
        public:
            BinaryOpNode(std::shared_ptr<ExprNode> leftOperand, int operatorCode, std::shared_ptr<ExprNode> rightOperand, int line, int col);
            static void printOp(int op);
            void unparse(int indent = 0);
            void checkSemantics();
            llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);
            quant_c_ast::QuantumValue quantum_codegen(QuantumContext& context);

            static bool classof(const ASTNode * node);

        // Members
            std::shared_ptr<ExprNode> leftOperand;
            std::shared_ptr<ExprNode> rightOperand;
            int operatorCode; // Token code of the operator
    }; // class BinaryOpNode

    class UnaryOpNode : public ExprNode {
        public:
            UnaryOpNode(std::shared_ptr<OperatorNode> op, std::shared_ptr<ExprNode> operand, int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);
            quant_c_ast::QuantumValue quantum_codegen(QuantumContext& context);

            static bool classof(const ASTNode * node);

        // Members
            std::shared_ptr<ExprNode> operand;
            std::shared_ptr<OperatorNode> opObject;
    }; // class UnaryOpNode

    class ConditionalExprNode : public ExprNode {
        public:
            ConditionalExprNode(std::shared_ptr<ExprNode> condition,
                            std::shared_ptr<ExprNode> lExpr,
                            std::shared_ptr<ExprNode> rExpr,
                            int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
        // Members
            std::shared_ptr<ExprNode> condition;
            std::shared_ptr<ExprNode> lExpr;
            std::shared_ptr<ExprNode> rExpr;
    }; // class ConditionalExprNode

    class AsgExprNode : public ExprNode {
        public:
            AsgExprNode(std::shared_ptr<ExprNode> lvalue, int operatorCode, std::shared_ptr<ExprNode> rvalue, int line, int col);
            static void printOp(int op);
            void unparse(int indent = 0);
            void checkSemantics();
            llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);
            quant_c_ast::QuantumValue quantum_codegen(QuantumContext& context);

            static bool classof(const ASTNode * node);

        // Members
            std::shared_ptr<ExprNode> lvalue;
            std::shared_ptr<ExprNode> rvalue;
            int assignOp;
    }; // class AsgExprNode

    class CastNode : public ExprNode {
        public:
            CastNode(std::shared_ptr<TypeNameNode> resultType, std::shared_ptr<ExprNode> operand, int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);

        // Members
            std::shared_ptr<ExprNode> operand;
            std::shared_ptr<TypeNameNode> resultType;
    }; // class CastNode

    class CallExprNode : public ExprNode {
        public:
            CallExprNode(std::shared_ptr<ExprNode> funcExpr, std::vector<std::shared_ptr<ExprNode>> funcArgs, int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);
            quant_c_ast::QuantumValue quantum_codegen(QuantumContext& context);

            static bool classof(const ASTNode * node);

        // Members
            std::shared_ptr<ExprNode> funcExpr;
            std::vector<std::shared_ptr<ExprNode>> funcArgs;
    }; // class CallExprNode

    class InitializerListNode : public ExprNode {
        public:
            InitializerListNode(
                std::vector<std::shared_ptr<ExprNode>> initalizerList,
                int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);
            quant_c_ast::QuantumValue quantum_codegen(QuantumContext& context);

            static bool classof(const ASTNode * node);
        // Members
            std::vector<std::shared_ptr<ExprNode>> initalizerList;

            // codegen
            bool isStruct = false;
    }; // class InitializerListNode

    class NullIdentNode : public IdentNode {
        public:
            NullIdentNode();
            bool isNull();
            void unparse(int indent = 0);
            static bool classof(const ASTNode * node);
    }; // class NullIdentNode

    class IndexExprNode : public ExprNode {
        public:
            IndexExprNode();
            IndexExprNode(std::shared_ptr<ExprNode> varName, std::shared_ptr<ExprNode> subscriptVal, int line, int col);
            IndexExprNode(std::shared_ptr<ExprNode> varName, int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);
            quant_c_ast::QuantumValue quantum_codegen(QuantumContext& context);

            static bool classof(const ASTNode * node);
        // Members
            std::shared_ptr<ExprNode> varName;
            std::shared_ptr<ExprNode> subscriptVal;
            bool genAddr = false;
    }; // class IndexExprNode

    class StrLitNode : public ExprNode {
        public:
            StrLitNode(std::string strval, std::string reducedStrval, int line, int col);
            void unparse(int indent = 0);
            llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
        // Members
            std::string strval;
            std::string reducedStrval;
    }; // class StrLitNode

    class IntLitNode : public ExprNode {
        public:
            IntLitNode();
            IntLitNode(std::intmax_t intval, quant_c_token::IntLitSize size, bool isSigned, int line, int col);
            void unparse(int indent = 0);
            llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);
            quant_c_ast::QuantumValue quantum_codegen(QuantumContext& context);


            static bool classof(const ASTNode * node);
        // Members
            std::intmax_t intval;
            quant_c_token::IntLitSize size;
            bool isSigned;
    }; // class IntLitNode

    class FloatLitNode : public ExprNode {
        public:
            FloatLitNode();
            FloatLitNode(long double val, quant_c_token::FloatLitSize size, int line, int col);
            void unparse(int indent = 0);
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);

        // Members
            long double floatValue;
            quant_c_token::FloatLitSize size;
    };

    class CharLitNode : public ExprNode {
        public:
            CharLitNode(char charval, int line, int col);
            std::string printCharValue();
            void unparse(int indent = 0);
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);

        // Members
            char charval;
    }; // class CharLitNode

    class TrueNode : public ExprNode {
        public:
            TrueNode(int line, int col);
            void unparse(int indent = 0);
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
    }; // class TrueNode

    class FalseNode : public ExprNode {
        public:
            FalseNode(int line, int col);
            void unparse(int indent = 0);
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
    }; // class FalseNode

    class OperatorNode : public ExprNode {
        public:
            OperatorNode(ASTNodeKind nodeKind);
            OperatorNode(ASTNodeKind nodeKind, int line, int col);
            static bool classof(const ASTNode * node);
    }; // class OperatorNode


    class UnaryOperatorNode : public OperatorNode {
        public:
            UnaryOperatorNode(int opCode);
            UnaryOperatorNode(int opCode, bool post, int line, int col);
            void unparse(int indent = 0);
            static bool classof(const ASTNode * node);

            // Members
            int opCode;
            bool isPostfix;
    }; // class UnaryOperatorNode


    class QuantumUnaryOperatorNode : public OperatorNode {
        public:
            enum class QuantumUnaryOp {
                IDENTITY,
                X_GATE,
                Y_GATE,
                Z_GATE,
                HADAMARD,
                PHASE,
                S_GATE,
                T_GATE,
                RX,
                RY,
                RZ,
                MEASURE
            };

            QuantumUnaryOperatorNode(QuantumUnaryOp op, int line, int col);
            QuantumUnaryOperatorNode(bool invert, QuantumUnaryOp op,
                std::vector<std::shared_ptr<QuantumUnaryPostfixNode>> modifiers,
                int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            quant_c_ast::QuantumValue quantum_codegen(QuantumContext& context);

            static bool classof(const ASTNode * node);

            // Members
            QuantumUnaryOp opCode;
            bool isInverted;
            std::vector<std::shared_ptr<QuantumUnaryPostfixNode>> opMods;
    }; // class QuantumUnaryOperatorNode

    class QuantumUnaryPostfixNode : public OperatorNode {
        public:
            QuantumUnaryPostfixNode(IdentNode id,int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            quant_c_ast::QuantumValue quantum_codegen(QuantumContext& context);

            static bool classof(const ASTNode * node);

            // Members
            IdentNode id;
    }; // class QuantumUnaryPostfixNode

    // end ExprNode subclasses


    /* ArgDeclNode & subclasses */
    class ArgDeclNode : public ASTNode {
        public:
            typedef std::vector<std::shared_ptr<TypeNode>> DeclSpecifierList;
            ArgDeclNode(ASTNodeKind nodeKind, DeclSpecifierList declSpecs);
            ArgDeclNode(ASTNodeKind nodeKind, DeclSpecifierList declSpecs,
                        int line, int col);

            type_info::Type generateArgType();
            static bool classof(const ASTNode * node);
        // Members
            DeclSpecifierList declSpecs;
            type_info::Type argType;
    }; // class ArgDeclNode

    /*
    class ArrayArgDeclNode : public ArgDeclNode {
        public:
            ArrayArgDeclNode(IdentNode argName, std::shared_ptr<TypeNode> elementType, int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);
        // Members
            std::shared_ptr<TypeNode> elementType;
    }; // class ArrayArgDeclNode

    class ValArgDeclNode : public ArgDeclNode {
        public:
            ValArgDeclNode(IdentNode argName, std::shared_ptr<TypeNode> argType, int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);
        // Members
            std::shared_ptr<TypeNode> argType;
    }; // class ValArgDeclNode
    */

    class NamedArgDeclNode : public ArgDeclNode {
        public:
            NamedArgDeclNode(DeclSpecifierList declSpecs,
                             std::shared_ptr<DeclaratorNode> decl,
                             int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
            // llvm::Type* get_type();
        // Members
            std::shared_ptr<DeclaratorNode> decl;
            symbol_table::SymbolInfo id_info;
    }; // class NamedArgDeclNode

    class AbstractArgDeclNode : public ArgDeclNode {
        public:
            AbstractArgDeclNode(DeclSpecifierList declSpecs,
                                std::shared_ptr<AbstractDeclaratorNode> decl,
                                int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
            // llvm::Type* get_type();
        // Members
            std::shared_ptr<AbstractDeclaratorNode> decl;
    }; // class AbstractArgDeclNode

    class TypeArgDeclNode : public ArgDeclNode {
        public:
            TypeArgDeclNode(DeclSpecifierList declSpecs,
                            int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
            // llvm::Type* get_type();
    }; // class TypeArgDeclNode

    // end ArgDeclNode subclasses



    /* StmtNode & subclasses (part 1) */
    class StmtNode : public ASTNode {
        public:
            StmtNode(ASTNodeKind nodeKind);
            StmtNode(ASTNodeKind nodeKind, int line, int col);

            static bool classof(const ASTNode * node);
            static std::shared_ptr<NullStmtNode> NULL_VAL;
    }; // class StmtNode

    class NullStmtNode : public StmtNode {
        public:
            NullStmtNode();
            bool isNull();

            static bool classof(const ASTNode * node);
    };

    class BlockNode : public StmtNode {
        public:
            static const bool returnRequired = true;
            static const bool returnNotNeccesary = false;
            static std::shared_ptr<NullBlockNode> NULL_VAL;
            BlockNode();
            BlockNode(std::vector<std::shared_ptr<DeclNode>> decls, std::vector<std::shared_ptr<StmtNode>> stmts, int line, int col);
            void setGivenScope(bool value);
            void setReturnStatus(bool status, type_info::Type type);
            void setReturnStatus(bool status);
            void setIsFuncBody();
            bool doesContainReturn();
            bool doesAlwaysReachReturn();
            void unparse(int indent = 0);
            void checkSemantics();
            llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);
            quant_c_ast::QuantumValue quantum_codegen(QuantumContext& context);

            static bool classof(const ASTNode * node);

        // Members
            std::vector<std::shared_ptr<DeclNode>> decls;
            std::vector<std::shared_ptr<StmtNode>> stmts;

            // Semantic checks
             // Flag for if node needs to create a new scope when checking semantics
            bool givenScope;
            // Flag for block containing a return statement
            bool containsReturn;
            // Flag for if all control paths lead to return
            bool alwaysReachesReturn;
            // Flag to indicate that this block is the body of a function
            bool isFuncBody;
            // Value to indicate constraints on containing a return value
            bool returnStatus;
            // Type to check for if this block must contain a return statement
            type_info::Type requiredReturnType;

            // Codegen
            // Parent (ex. function or instruction) that this block belongs to
            llvm::Value* parentCode;
            // Block that break statements need to jump to
            llvm::Value* succesorBlock;
    }; // class BlockNode

    class NullBlockNode : public BlockNode {
        public:
            NullBlockNode();
            bool isNull();
            void unparse(int indent = 0);

            static bool classof(const ASTNode * node);
    }; // class NullBlockNode

    // end StmtNode subclasses (part 1)

    /* StmtNode & subclasses (part 2) */
    /*
    class AsgNode : public StmtNode {
        public:
            AsgNode(IndexExprNode target, std::shared_ptr<ExprNode> source, int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);

        // Members
            IndexExprNode target;
            std::shared_ptr<ExprNode> source;
    }; // class AsgNode
    */

    class IfThenNode : public StmtNode {
        public:
            IfThenNode(std::shared_ptr<ExprNode> condition, std::shared_ptr<StmtNode> thenPart, std::shared_ptr<StmtNode> elsePart, int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);

        // Members
            std::shared_ptr<ExprNode> condition;
            std::shared_ptr<StmtNode> thenPart;
            std::shared_ptr<StmtNode> elsePart;
    }; // class IfThenNode

    class ForNode : public StmtNode {
        public:
            ForNode(std::shared_ptr<StmtNode> initializer_stmt, std::shared_ptr<StmtNode> condition_stmt,
            std::shared_ptr<ExprNode> update_expr, std::shared_ptr<StmtNode> body, int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);

        // Members
            std::shared_ptr<StmtNode> initializer_stmt;
            std::shared_ptr<StmtNode> condition_stmt;
            std::shared_ptr<ExprNode> update_expr;
            std::shared_ptr<StmtNode> body;
    }; // class ForNode

    class WhileNode : public StmtNode {
        public:
            WhileNode(std::shared_ptr<ExprNode> condition, std::shared_ptr<StmtNode> loopBody, bool is_do_while, int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);

        // Members
            std::shared_ptr<ExprNode> condition;
            std::shared_ptr<StmtNode> body;
            bool is_do_while;
    }; // class WhileNode

    class BreakNode : public StmtNode {
        public:
            BreakNode(int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);

        // Members
            // Codegen
            llvm::BasicBlock* breakToLocation = nullptr;


    }; // class BreakNode

    class ContinueNode : public StmtNode {
        public:
            ContinueNode(int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
    }; // class ContinueNode

    class ReturnNode : public StmtNode {
        public:
            ReturnNode(std::shared_ptr<ExprNode> returnVal);
            ReturnNode(std::shared_ptr<ExprNode> returnVal, int line, int col);
            void unparse(int indent = 0);
            void setExpectedType(type_info::Type type);
            void checkSemantics();
            llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);
            quant_c_ast::QuantumValue quantum_codegen(QuantumContext& context);

            static bool classof(const ASTNode * node);
        // Members
            std::shared_ptr<ExprNode> returnVal;
            type_info::Type expectedType;
    }; // class ReturnNode

    class ExprStmtNode : public StmtNode {
        public:
            ExprStmtNode(std::shared_ptr<ExprNode> expression, int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);
            quant_c_ast::QuantumValue quantum_codegen(QuantumContext& context);

            static bool classof(const ASTNode * node);
        // Members
            std::shared_ptr<ExprNode> expression;
    }; // class ExprStmtNode

    class LabelStmtNode : public StmtNode {
        public:
            LabelStmtNode(IdentNode label, std::shared_ptr<StmtNode> stmt, int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
        // Members
            IdentNode label;
            std::shared_ptr<StmtNode> stmt;
    }; // class LabelStmtNode

    class SwitchNode : public StmtNode {
        public:
            SwitchNode(std::shared_ptr<ExprNode> condition, std::shared_ptr<StmtNode> body, int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);

        // Members
            std::shared_ptr<ExprNode> condition;
            std::shared_ptr<StmtNode> body;

            // Semantics checks
            bool allBranchesReturn = false;
            bool explicitDefault = false;
    };

    class CaseStmtNode : public StmtNode {
        public:
            CaseStmtNode(std::shared_ptr<ExprNode> case_val, std::shared_ptr<StmtNode> stmt, int line, int col);
            CaseStmtNode(std::shared_ptr<StmtNode> stmt, int line, int col);
            void unparse(int indent = 0);
            void setReturnStatus(bool status, type_info::Type type);
            void setReturnStatus(bool status);
            void checkSemantics();
            llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
        // Members
            std::shared_ptr<ExprNode> case_val;
            std::shared_ptr<StmtNode> stmt;
            bool is_default;

            // Semantics checks
            // Value to indicate constraints on containing a return value
            bool returnStatus;
            // Type to check for if this block must contain a return statement
            type_info::Type requiredReturnType;
            // Flag for if this case always reaches a return
            bool alwaysReturns = false;
            // Flag for if this case is not the default but adjacent to it
            bool leadsToDefault = false;
    }; // class CaseStmtNode

    class GotoNode : public StmtNode {
        public:
            GotoNode(std::shared_ptr<IdentNode> label, int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);

        // Members
            std::shared_ptr<IdentNode> label;
    };



    // end rest of StmtNode subclasses


    /* DeclNode & subclasses */
    class DeclNode : public ASTNode {
        // abstract superclass; only subclasses are actually created
        public:
            DeclNode(ASTNodeKind nodeKind);
            DeclNode(ASTNodeKind nodeKind, int line, int col);

            static bool classof(const ASTNode * node);
    }; // class DeclNode

    class DeclarationNode : public DeclNode {
        public:
            DeclarationNode(std::vector<std::shared_ptr<TypeNode>> declSpec,
                  std::vector<std::shared_ptr<InitDeclaratorNode>> initList,
                  int line, int col);
            void unparse(int indent = 0);
            type_info::Type generate_type();
            void checkSemantics();
            llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);
            quant_c_ast::QuantumValue quantum_codegen(QuantumContext& context);

            static bool classof(const ASTNode * node);
        // Members
            std::vector<std::shared_ptr<TypeNode>> declSpec;
            std::vector<std::shared_ptr<InitDeclaratorNode>> initList;
            std::vector<symbol_table::SymbolInfo> id_info;
            bool isGlobalDec = false;
            bool isWithinQuantumBlock = false;
    }; // class DeclarationNode

    class InitDeclaratorNode : public DeclNode {
        public:
            InitDeclaratorNode(std::shared_ptr<DeclaratorNode> decl,
                               int line, int col);
            InitDeclaratorNode(std::shared_ptr<DeclaratorNode> decl,
                               std::shared_ptr<ExprNode> initializer,
                               int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);
            quant_c_ast::QuantumValue quantum_codegen(QuantumContext& context);

            static bool classof(const ASTNode * node);
        // Members
            std::shared_ptr<DeclaratorNode> decl;
            std::shared_ptr<ExprNode> initializer;
            bool hasInitalizer;
            type_info::Type initializerType;
            std::string id_name;
        // Codegen members
            type_info::TypeQualityVal quality = type_info::TypeQualityVal::NONE;

    }; // class InitDeclaratorNode

    /*
    class VariableDeclNode : public DeclNode {
        public:
            VariableDeclNode(ASTNodeKind nodeKind);
            VariableDeclNode(ASTNodeKind nodeKind, int line, int col);

            static bool classof(const ASTNode * node);
    }; // class VariableDeclNode

    class VarDeclNode : public VariableDeclNode {
        public:

            VarDeclNode(IdentNode varName, std::shared_ptr<TypeNode> varType, std::shared_ptr<ExprNode> initValue, int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode* node);
        // Members
            IdentNode varName;
            std::shared_ptr<TypeNode> varType;
            std::shared_ptr<ExprNode> initValue;
    }; // class VarDeclNode

    class ConstDeclNode : public VariableDeclNode {
        public:
            ConstDeclNode(IdentNode constName, std::shared_ptr<ExprNode> constValue, int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);

        // Members
            IdentNode constName;
            std::shared_ptr<ExprNode> constValue;
    }; // class ConstDeclNode

    class ArrayDeclNode : public VariableDeclNode {
        public:
            ArrayDeclNode(IdentNode arrayName, std::shared_ptr<TypeNode> elementType, IntLitNode arraySize, int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);

        // Members
            IdentNode arrayName;
            std::shared_ptr<TypeNode> elementType;
            IntLitNode arraySize;
    }; // class ArrayDeclNode
    */

    class FuncDefNode : public DeclNode {
        public:
            typedef std::vector<std::shared_ptr<TypeNode>> DeclSpecifierList;
            FuncDefNode(DeclSpecifierList returnType, std::shared_ptr<DeclaratorNode> declarator, std::shared_ptr<BlockNode> body, int line, int col);
            void unparse(int indent = 0);
            type_info::Type generate_return_type();
            void checkSemantics();
            llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);

        // Members
            DeclSpecifierList returnType;
            std::shared_ptr<DeclaratorNode> declarator;
            std::shared_ptr<BlockNode> body;

            symbol_table::SymbolInfo id_info;
    }; // class FuncDefNode

    class DeclaratorNode : public DeclNode {
        public:
            typedef std::vector<std::shared_ptr<PointerNode>> PointerList;
            DeclaratorNode(PointerList pointer_spec,
                           std::shared_ptr<DirectDeclaratorNode> direct_decl,
                           int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);
            quant_c_ast::QuantumValue quantum_codegen(QuantumContext& context);

            static bool classof(const ASTNode * node);
        // Members
            PointerList pointer_spec;
            std::shared_ptr<DirectDeclaratorNode> direct_decl;
            std::string id_name;

    }; // class DeclaratorNode

    class DirectDeclaratorNode : public DeclNode {
        public:
            DirectDeclaratorNode(ASTNodeKind nodeKind, int line, int col);

            static bool classof(const ASTNode * node);
        // Members
            std::string id_name;
    }; // class DirectDeclaratorNode

    class IdentifierDeclaratorNode : public DirectDeclaratorNode {
        public:
            IdentifierDeclaratorNode(std::string name,
                                     int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
    }; // class IdentifierDeclaratorNode

    class ParensDeclaratorNode : public DirectDeclaratorNode {
        public:
            ParensDeclaratorNode(std::shared_ptr<DeclaratorNode> decl,
                                 int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);
            quant_c_ast::QuantumValue quantum_codegen(QuantumContext& context);

            static bool classof(const ASTNode * node);
        // Members
            std::shared_ptr<DeclaratorNode> decl;
    }; // class ParensDeclaratorNode

    class ArrayDeclaratorNode : public DirectDeclaratorNode {
        public:
            ArrayDeclaratorNode(std::shared_ptr<DirectDeclaratorNode> decl,
                                std::shared_ptr<ExprNode> size,
                                 int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
        // Members
            std::shared_ptr<DirectDeclaratorNode> decl;
            std::shared_ptr<ExprNode> size;
    }; // class ArrayDeclaratorNode

    class FunctionDeclaratorNode : public DirectDeclaratorNode {
        public:
            typedef std::vector<std::shared_ptr<ArgDeclNode>> ArgumentList;
            FunctionDeclaratorNode(std::shared_ptr<DirectDeclaratorNode> decl,
                                   ArgumentList arguments,
                                   bool isVarArg,
                                   int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);

            static bool classof(const ASTNode * node);
        // Members
            std::shared_ptr<DirectDeclaratorNode> decl;
            ArgumentList arguments;
            bool isVarArg;

            std::vector<std::string> argNames;
    }; // class FunctionDeclaratorNode

    class DestructureDeclaratorNode : public DirectDeclaratorNode {
        public:
            typedef std::vector<std::shared_ptr<quant_c_token::IdentifierToken>>
                IdentList;
            DestructureDeclaratorNode(IdentList ids,
                                   int line, int col);
            void unparse(int indent = 0);
            void checkSemantics();
            // llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);
            // quant_c_ast::QuantumValue quantum_codegen(QuantumContext& context);

            static bool classof(const ASTNode * node);
        // Members
            IdentList identifiers;

    }; // class DestructureDeclaratorNode


    class AbstractDeclaratorNode : public DeclNode {
        public:
            typedef std::vector<std::shared_ptr<PointerNode>> PointerList;
            AbstractDeclaratorNode(ASTNodeKind nodeKind);
            AbstractDeclaratorNode(PointerList pointer_spec,
                           std::shared_ptr<DirectAbstractDeclaratorNode> direct_decl,
                           int line, int col);
            void unparse(int indent = 0);
            type_info::Type generateAbstractType();
            void checkSemantics();

            static bool classof(const ASTNode * node);
            static std::shared_ptr<NullAbstractNode> NULL_VAL;
        // Members
            PointerList pointer_spec;
            std::shared_ptr<DirectAbstractDeclaratorNode> direct_decl;
    }; // class AbstractDeclaratorNode

    class NullAbstractNode : public AbstractDeclaratorNode {
        public:
            NullAbstractNode();
            bool isNull();
            void unparse(int indent = 0);
            type_info::Type generateAbstractType();

            static bool classof(const ASTNode * node);
    };

    class DirectAbstractDeclaratorNode : public DeclNode {
        public:
            DirectAbstractDeclaratorNode(ASTNodeKind nodeKind);
            type_info::Type generateDirectAbstractType();
            static std::shared_ptr<NullDirectAbstractNode> NULL_VAL;

            static bool classof(const ASTNode * node);
    };

    class NullDirectAbstractNode : public DirectAbstractDeclaratorNode {
        public:
            NullDirectAbstractNode();
            type_info::Type generateDirectAbstractType();
            bool isNull();
            void unparse(int indent = 0);

            static bool classof(const ASTNode * node);
    };

    // end DeclNode subclasses

    /* TranslationUnitNode */
    class TranslationUnitNode : public ASTNode {
        public:
            TranslationUnitNode();
            TranslationUnitNode(std::vector<std::shared_ptr<DeclNode>> declList, int line, int col);
            void setName(std::string name);
            void unparse(int indent = 0);
            void checkSemantics();
            llvm::Value* codegen(llvm::Value* parent = nullptr, llvm::Type* typeInfo = nullptr);
            bool isSemanticsCorrect();

            static bool classof(const ASTNode * node);
        // Members
            std::vector<std::shared_ptr<DeclNode>> declList;
            std::string module_name;
    }; // class TranslationUnitNode


}
