#pragma once
#include <string>
#include <unordered_map>
#include <cstdint>
#include <exception>

namespace quant_c_token {

    class InvalidTokenException: public std::exception {
        std::string bad_token;
        int linenum;
        int colnum;

        public:
            InvalidTokenException(std::string token, int line, int col):
                bad_token{token}, linenum{line}, colnum{col} {}
            const char * what() const noexcept override;
    };

    class Token {
        private:
            // The line number where the Token is located in a piece of text
            int linenum;

            // The column number where the Token is located in a piece of text
            int colnum;

        public:
            Token() = default;
            Token(int line, int col);
            int get_line() { return this->linenum; };
            int get_col()  { return this->colnum; };
    };

    class BoolLitToken : public Token {
        private:
            // The boolean value of a scanned boolean literal
            bool boolValue;

        public:
            BoolLitToken() = default;
            BoolLitToken(bool boolValue, int line, int col);
            bool get_value();
    };

    class CharLitToken : public Token {
        private:
            // The character value of a valid character literal
            char charValue;

        public:
            CharLitToken() = default;
            CharLitToken(char val, int line, int col);
            char get_value();

            // Produce a printable version of the character value associated with this token.
            std::string printCharValue();
    };

    class ErrorToken : public Token {
        private:
            // The erroneous token string
            std::string badToken;

        public:
            ErrorToken() = default;
            ErrorToken(std::string value, int line, int col);
            std::string get_value();
    };

    class IdentifierToken : public Token {
        private:
            // The identifier name
            std::string name;

        public:
            IdentifierToken() = default;
            IdentifierToken(std::string name, int line, int col);
            std::string get_value();
    };

    enum class IntLitSize {
        INT,
        LONG,
        LONGLONG
    };

    class IntLitToken : public Token {
        // The integer value of an integer literal
        private:
            std::intmax_t intValue;
            IntLitSize size;
            bool isSigned;

        public:
            IntLitToken() = default;
            IntLitToken(std::intmax_t val, IntLitSize size, bool isSigned, int line, int col);
            std::intmax_t get_value();
            IntLitSize get_size();
            bool is_signed();

    };

    enum class FloatLitSize {
        FLOAT,
        DOUBLE,
        LONGDOUBLE
    };

    class FloatLitToken : public Token {
        // The integer value of an integer literal
        private:
            long double floatValue;
            FloatLitSize size;

        public:
            FloatLitToken() = default;
            FloatLitToken(long double val, FloatLitSize size, int line, int col);
            double get_value();
            FloatLitSize get_size();
    };

    class TypeNameToken : public Token {
        private:
            // The type name
            std::string name;

        public:
            TypeNameToken() = default;
            TypeNameToken(std::string name, int line, int col);
            std::string get_value();
    };

    class StringLitToken : public Token {
        private:
            // Full text of string literal, including quotes & escapes
            std::string string;

            // Minimized text of string literal,  removing quotes and replacing escapes
            std::string reducedString;

            // Initializer for static map
            static std::unordered_map<std::string, std::string> _static_init();

        public:
            // Map of substitutions used to minimize a string literal
            static std::unordered_map<std::string, std::string> substitutions;

            StringLitToken() = default;
            StringLitToken(std::string text, int line, int col);
            std::string get_value(bool raw_string = true);
    };

}
